@@ -1,6 +1,7 @@
 """
 LazyArray
 """
+import atexit
 import cPickle
 import logging
 import os
@@ -16,13 +17,13 @@ logger = logging.getLogger(__name__)
 class InferenceError(Exception):
     """Information about a lazily-evaluated quantity could not be inferred"""
 
+
 class UnknownShape(InferenceError):
     """Shape could not be inferred"""
 
+
 def is_int_idx(idx):
-    #XXX: add numpy int types
-    return isinstance(idx,
-            (int,))
+    return isinstance(idx, (int, np.integer))
 
 
 def is_larray(thing):
@@ -102,6 +103,9 @@ class lmap(larray):
         """
         ragged = kwargs.pop('ragged', False)
         f_map = kwargs.pop('f_map', None)
+        if f_map is None:
+            f_map = getattr(fn, 'f_map', None)
+
         if kwargs:
             raise TypeError('unrecognized kwarg', kwargs.keys())
 
@@ -271,12 +275,16 @@ class loop(larray):
 
 
 class reindex(larray):
+    """
+    Lazily re-index list-like `obj` so that
+    `self[i]` means `obj[imap[i]]`
+    """
     def __init__(self, obj, imap):
         self.obj = obj
         self.imap = np.asarray(imap)
         if 'int' not in str(self.imap.dtype):
             #XXX: diagnostic info
-            raise TypeError()
+            raise TypeError(imap.dtype)
 
     def __len__(self):
         return len(self.imap)
@@ -386,10 +394,11 @@ class CacheMixin(object):
         if batchsize <= 0:
             raise ValueError('non-positive batch size')
         if batchsize == 1:
-            for i in xrange(self.shape[0]):
+            for i in xrange(len(self)):
                 self[i]
         else:
-            while i < self.shape[0]:
+            i = 0
+            while i < len(self):
                 self[i:i + batchsize]
                 i += batchsize
 
@@ -418,7 +427,7 @@ class CacheMixin(object):
         return [self.obj]
 
     def __getitem__(self, item):
-        if isinstance(item, (int, np.int)):
+        if isinstance(item, (int, np.integer)):
             if self._valid[item]:
                 return self._data[item]
             else:
@@ -433,13 +442,56 @@ class CacheMixin(object):
             assert v.ndim == 1
             if np.all(v):
                 return self._data[item]
-            # otherwise at least some of the requested elements must be
-            # computed
-            # XXX: this brute-force recomputes all of the requested elements
+
+            # -- Quick and dirty, yes.
+            # -- Accurate, ?
+            try:
+                list(item)
+                is_int_list = True
+            except:
+                is_int_list = False
+
+            if np.sum(v) == 0:
+                # -- we need to re-compute everything in item
+                sub_item = item
+            elif is_int_list:
+                # -- in this case advanced indexing has been used
+                #    and only some of the elements need to be recomputed
+                assert self._valid.max() <= 1
+                item = np.asarray(item)
+                assert 'int' in str(item.dtype)
+                sub_item = item[v == 0]
+            elif isinstance(item, slice):
+                # -- in this case slice indexing has been used
+                #    and only some of the elements need to be recomputed
+                #    so we are converting the slice to an int_list
+                idxs_of_v = np.arange(len(self._valid))[item]
+                sub_item = idxs_of_v[v == 0]
+            else:
+                sub_item = item
+
             self.rows_computed += v.sum()
-            values = self.obj[item]
-            self._valid[item] = 1
-            self._data[item] = values
+            sub_values = self.obj[sub_item]  # -- retrieve missing elements
+            self._valid[sub_item] = 1
+            try:
+                self._data[sub_item] = sub_values
+            except:
+                logger.error('data dtype %s' % str(self._data.dtype))
+                logger.error('data shape %s' % str(self._data.shape))
+
+                logger.error('sub_item str %s' % str(sub_item))
+                logger.error('sub_item type %s' % type(sub_item))
+                logger.error('sub_item len %s' % len(sub_item))
+                logger.error('sub_item shape %s' % getattr(sub_item, 'shape',
+                    None))
+
+                logger.error('sub_values str %s' % str(sub_values))
+                logger.error('sub_values type %s' % type(sub_values))
+                logger.error('sub_values len %s' % len(sub_values))
+                logger.error('sub_values shape %s' % getattr(sub_values, 'shape',
+                    None))
+                raise
+            assert np.all(self._valid[item])
             return self._data[item]
 
 
@@ -461,7 +513,7 @@ class cache_memory(CacheMixin, larray):
     def clone(self, given):
         return self.__class__(obj=given_get(given, self.obj))
 
-    
+
 class cache_memmap(CacheMixin, larray):
     """
     Provide a lazily-filled cache of a larray (obj) via a memmap file
@@ -475,7 +527,7 @@ class cache_memmap(CacheMixin, larray):
 
     ROOT = os.path.join(get_data_home(), 'memmaps')
 
-    def __init__(self, obj, name, basedir=None, msg=None):
+    def __init__(self, obj, name, basedir=None, msg=None, del_atexit=False):
         """
         If new files are created, then `msg` will be written to README.msg
         """
@@ -506,8 +558,10 @@ class cache_memmap(CacheMixin, larray):
                 logger.warn("Problem re-using memmap: dtype/shape mismatch")
                 logger.info('Creating memmap %s with dtype %s, shape %s' % (
                         data_path,
-                        str(dtype),
+                        str(obj.dtype),
                         str(obj.shape)))
+                dtype = obj.dtype
+                shape = obj.shape
         except IOError:
             dtype = obj.dtype
             shape = obj.shape
@@ -545,13 +599,15 @@ class cache_memmap(CacheMixin, larray):
 
         self.rows_computed = 0
 
+        if del_atexit:
+            atexit.register(self.delete_files)
+
     def delete_files(self):
+        logger.info('deleting cache_memmap at %s' % self.dirname)
         subprocess.call(['rm', '-Rf', self.dirname])
 
     def clone(self, given):
         raise NotImplementedError()
-        return self.__class__(
-            obj=given_get(given, self.obj),
-            dirname=self.dirname + '_clone')
-    # XXX: What if you clone more than once? This implementation would
-    #      cause interference
+        # XXX: careful to ensure that any instance can be cloned multiple
+        # times, and the clones can themselves be cloned recursively.
+
